(*=============================================================================|
|  PROJECT SNAP7                                                         1.4.0 |
|==============================================================================|
|  Copyright (C) 2013, 2015 Davide Nardella                                    |
|  All rights reserved.                                                        |
|==============================================================================|
|  SNAP7 is free software: you can redistribute it and/or modify               |
|  it under the terms of the Lesser GNU General Public License as published by |
|  the Free Software Foundation, either version 3 of the License, or           |
|  (at your option) any later version.                                         |
|                                                                              |
|  It means that you can distribute your commercial software linked with       |
|  SNAP7 without the requirement to distribute the source code of your         |
|  application and without the requirement that your application be itself     |
|  distributed under LGPL.                                                     |
|                                                                              |
|  SNAP7 is distributed in the hope that it will be useful,                    |
|  but WITHOUT ANY WARRANTY; without even the implied warranty of              |
|  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               |
|  Lesser GNU General Public License for more details.                         |
|                                                                              |
|  You should have received a copy of the GNU General Public License and a     |
|  copy of Lesser GNU General Public License along with Snap7.                 |
|  If not, see  http://www.gnu.org/licenses/                                   |
|=============================================================================*)
unit snap7;

// Snap7 library contains three objects: the Client, the Server and the Partner.
// The partner that requests the connection is named active, the one that accept the connection is named passive.
// The communication is not fully asynchronous, i.e. there is no interrupt mechanism that says to the receiver that a packet is incoming:
//  a partner to receive a packet, must be listening via a Block Recv function. (see Par_BRecv)

{$IFDEF FPC}
  {$mode objfpc}
{$ENDIF}
{$H+}

{$DEFINE LOAD_DYNAMICALLY}

interface

const
{$IFDEF MSWINDOWS}
  libsnap7 = 'snap7.dll';
{$ELSE}
  libsnap7 = 'libsnap7.so';
{$ENDIF}

  // Client Connection Type
  CONNTYPE_PG    = $01;  // Connect to the PLC as a PG (the programming console)
  CONNTYPE_OP    = $02;  // Connect to the PLC as an OP (the Siemens HMI panel)
  CONNTYPE_BASIC = $03;  // Basic connection (a generic data transfer connection)

  // Area ID
  S7AreaDB   =	$84;
  // ...

  // Word Length
  S7WLBit     = $01;
  S7WLByte    = $02;
  S7WLChar    = $03;
  S7WLWord    = $04;
  // ...

  // CPU status
  S7CpuStatusUnknown = $00;
  S7CpuStatusRun     = $08;
  S7CpuStatusStop    = $04;

  // API Error codes
  //  00-15: Low 16 bits (4 nibbles) represent a TCP/IP error that is raised by the OS socket layer.
  //  16-19: The 5th nibble represent an ISO TCP error
  //  20-31: High three nibbles represent an S7 protocol error

  errIsoMask = $000F0000;
  errIsoConnect         = $00010000; // Connection error
  errIsoDisconnect      = $00020000; // Disconnect error
  errIsoInvalidPDU      = $00030000; // Bad format
  errIsoInvalidDataSize = $00040000; // Bad Datasize passed to send/recv : buffer is invalid
  errIsoNullPointer    	= $00050000; // Null passed as pointer
  errIsoShortPacket    	= $00060000; // A short packet received
  errIsoTooManyFragments= $00070000; // Too many packets without EoT flag
  errIsoPduOverflow    	= $00080000; // The sum of fragments data exceded maximum packet size
  errIsoSendPacket      = $00090000; // An error occurred during send
  errIsoRecvPacket      = $000A0000; // An error occurred during recv
  errIsoInvalidParams   = $000B0000; // Invalid TSAP params

  errCliMask = $FFF00000;
  errCliInvalidParams          = $00200000;
  errCliJobPending             = $00300000;
  errCliTooManyItems           = $00400000;
  errCliInvalidWordLen         = $00500000;
  errCliPartialDataWritten     = $00600000;
  errCliSizeOverPDU            = $00700000;
  errCliInvalidPlcAnswer       = $00800000;
  errCliAddressOutOfRange      = $00900000;
  errCliInvalidTransportSize   = $00A00000;
  errCliWriteDataSizeMismatch  = $00B00000;
  errCliItemNotAvailable       = $00C00000;
  errCliInvalidValue           = $00D00000;
  errCliCannotStartPLC         = $00E00000;
  errCliAlreadyRun             = $00F00000;
  errCliCannotStopPLC          = $01000000;
  errCliCannotCopyRamToRom     = $01100000;
  errCliCannotCompress         = $01200000;
  errCliAlreadyStop            = $01300000;
  errCliFunNotAvailable        = $01400000;
  errCliUploadSequenceFailed   = $01500000;
  errCliInvalidDataSizeRecvd   = $01600000;
  errCliInvalidBlockType       = $01700000;
  errCliInvalidBlockNumber     = $01800000;
  errCliInvalidBlockSize       = $01900000;
  errCliDownloadSequenceFailed = $01A00000;
  errCliInsertRefused          = $01B00000;
  errCliDeleteRefused          = $01C00000;
  errCliNeedPassword           = $01D00000;
  errCliInvalidPassword        = $01E00000;
  errCliNoPasswordToSetOrClear = $01F00000;
  errCliJobTimeout             = $02000000;
  errCliPartialDataRead        = $02100000;
  errCliBufferTooSmall         = $02200000;
  errCliFunctionRefused        = $02300000;
  errCliDestroying             = $02400000;
  errCliInvalidParamNumber     = $02500000;
  errCliCannotChangeParam      = $02600000;

type
  TS7Object = NativeUInt;

  TS7CpuInfo = packed record
    ModuleTypeName : packed array[0..32] of AnsiChar;
    SerialNumber   : packed array[0..24] of AnsiChar;
    ASName         : packed array[0..24] of AnsiChar;
    Copyright      : packed array[0..26] of AnsiChar;
    ModuleName     : packed array[0..24] of AnsiChar;
  end;
  PS7CpuInfo = ^TS7CpuInfo;

{$IFDEF LOAD_DYNAMICALLY}
var
  Cli_Create: function():TS7Object; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  Cli_Destroy: procedure(var Client : TS7Object); {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  Cli_SetConnectionType: function(Client : TS7Object; ConnectionType : word) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  Cli_ConnectTo: function(Client : TS7Object; Address : PAnsiChar; Rack, Slot : integer): integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_Disconnect: Disconnects “gracefully” the Client from the PLC. If Client parameter is a valid handle, this function returns always 0, it can be called safely multiple times.
  //  This function is called internally by Cli_Destroy() too.
  Cli_Disconnect: function(Client : TS7Object): integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_ReadArea is the main function to read data from a PLC. With it you can read DB, Inputs, Outputs, Merkers, Timers and Counters.
  Cli_ReadArea: function(Client : TS7Object; Area, DBNumber, Start, Amount, WordLen : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  Cli_WriteArea: function(Client : TS7Object; Area, DBNumber, Start, Amount, WordLen : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_DBRead internally calls Cli_ReadArea() with Area = S7AreaDB and WordLen = S7WLByte
  Cli_DBRead: function(Client : TS7Object; DBNumber, Start, Size : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_DBWrite internally calls Cli_WriteArea() with Area = S7AreaDB and WordLen = S7WLByte
  Cli_DBWrite: function(Client : TS7Object; DBNumber, Start, Size : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  Cli_GetCpuInfo: function(Client : TS7Object; pUsrData : PS7CpuInfo) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  Cli_GetPlcStatus: function(Client : TS7Object; var Status : integer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_GetLastError: 0=The function was accomplished with no errors
  Cli_GetLastError: function(Client : TS7Object; var LastError : integer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_ErrorText returns a textual explanation of a given error number.
  Cli_ErrorText: function (Error : integer; Text : PAnsiChar; TextLen : integer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
  // Cli_GetConnected: IsConnected is 0 if the client is not connected otherwise contains an integer !=0.
  // (returns value of client class field ->Connected; if error LastTcpError==WSAECONNRESET (10054) occured during recv (but not send) then Connected becomes 0 (See: TMsgSocket::WaitForData, TMsgSocket::Receive, TMsgSocket::RecvPacket))
  Cli_GetConnected: function(Client : TS7Object; var IsConnected: integer): integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF};
procedure LoadSnap7();
procedure UnloadSnap7;
{$ELSE}
function Cli_Create: TS7Object; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
procedure Cli_Destroy(var Client : TS7Object); {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_SetConnectionType(Client : TS7Object; ConnectionType : word) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_ConnectTo(Client : TS7Object; Address : PAnsiChar; Rack, Slot : integer): integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_Disconnect(Client : TS7Object): integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_ReadArea(Client : TS7Object; Area, DBNumber, Start, Amount, WordLen : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_WriteAreafunction(Client : TS7Object; Area, DBNumber, Start, Amount, WordLen : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_DBRead(Client : TS7Object; DBNumber, Start, Size : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_DBWrite(Client : TS7Object; DBNumber, Start, Size : integer; pUsrData : pointer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_GetCpuInfo(Client : TS7Object; pUsrData : PS7CpuInfo) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_GetPlcStatus(Client : TS7Object; var Status : integer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_GetLastError(Client : TS7Object; var LastError : integer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_ErrorText(Error : integer; Text : PAnsiChar; TextLen : integer) : integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
function Cli_GetConnected(Client : TS7Object; var IsConnected: integer): integer; {$IFDEF MSWINDOWS}stdcall{$ELSE}cdecl{$ENDIF}; external libsnap7;
{$ENDIF}

function Cli_IsConnected(Client: TS7Object): boolean;

implementation

{$IFDEF LOAD_DYNAMICALLY}
uses
  SysUtils,{$IFDEF WINDOWS}Windows{$ELSE}DynLibs{$ENDIF};

var
  libsnap7Handle: {$IFDEF WINDOWS}THandle{$ELSE}TLibHandle{$ENDIF};
  RefCount: integer;

{$IFDEF WINDOWS}
function GetProcedureAddress(Lib: THandle; ProcName: PAnsiChar): Pointer; inline;
begin
  Result := Windows.GetProcAddress(Lib, ProcName);
end;
function UnloadLibrary(Lib: THandle): Boolean; inline;
begin
  Result := Windows.FreeLibrary(Lib);
end;
{$ENDIF}

procedure LoadSnap7();
begin
  Inc(RefCount);
  if RefCount = 1 then begin
    libsnap7Handle := LoadLibrary(libsnap7);
    if libsnap7Handle = 0 then begin
      RefCount := 0;
      raise EInOutError.CreateFmt('Can not load library "%s". Check your installation.'+LineEnding+'%s',
                                  [libsnap7, SysErrorMessage(GetLastOSError)]);
    end;

    pointer(Cli_Create) := GetProcedureAddress(libsnap7Handle,'Cli_Create');
    pointer(Cli_Destroy) := GetProcedureAddress(libsnap7Handle,'Cli_Destroy');
    pointer(Cli_SetConnectionType) := GetProcedureAddress(libsnap7Handle,'Cli_SetConnectionType');
    pointer(Cli_ConnectTo) := GetProcedureAddress(libsnap7Handle,'Cli_ConnectTo');
    pointer(Cli_Disconnect) := GetProcedureAddress(libsnap7Handle,'Cli_Disconnect');
    pointer(Cli_ReadArea) := GetProcedureAddress(libsnap7Handle,'Cli_ReadArea');
    pointer(Cli_WriteArea) := GetProcedureAddress(libsnap7Handle,'Cli_WriteArea');
    pointer(Cli_DBRead) := GetProcedureAddress(libsnap7Handle,'Cli_DBRead');
    pointer(Cli_DBWrite) := GetProcedureAddress(libsnap7Handle,'Cli_DBWrite');
    pointer(Cli_GetCpuInfo) := GetProcedureAddress(libsnap7Handle,'Cli_GetCpuInfo');
    pointer(Cli_GetPlcStatus) := GetProcedureAddress(libsnap7Handle,'Cli_GetPlcStatus');
    pointer(Cli_GetLastError) := GetProcedureAddress(libsnap7Handle,'Cli_GetLastError');
    pointer(Cli_ErrorText) := GetProcedureAddress(libsnap7Handle,'Cli_ErrorText');
    pointer(Cli_GetConnected) := GetProcedureAddress(libsnap7Handle,'Cli_GetConnected');
  end;
end;

procedure UnloadSnap7;
begin
  if RefCount > 0 then begin
    Dec(RefCount);
    if RefCount = 0 then
      UnloadLibrary(libsnap7Handle);
  end;
end;
{$ENDIF}

function Cli_IsConnected(Client: TS7Object): boolean;
var
  IsConnected : integer;
begin
  IsConnected:=0;
  if Cli_GetConnected(Client, IsConnected)=0 then
    Result:=IsConnected<>0
  else
    Result:=False;
end;

end.

